; zforth: the Pocket80 built-in language.              *AS/C/UNIX source code*
; Definitely "inspired" by False. Just weirder. Think "compressed FORTH".
; History:	v0.7 Now accepts long names for functions.
;		v0.8 lf has been abolished. Ported to CHAOS, badly, adding }.
;
; To do:	make it load the word area from a file and save it again
			
;;; I need this because I use a LD A,IXU at one point to check the stack.
	cpu Z80undoc
	
;;; Include some stuff.
	include chaos.i
		
; A few values that application programs don't need to know.
numstack equ 1FFEh		; Numbers stack.
recurselimit equ 18	; Maximum level of recursion. This is conservative.
	
	ORG appbase
	JR restart
	
;;; System variables.

exityn:	DB 0			; Becomes a 1 when we want to quit.
recurselevel:	DB 0		; The level of recursion we are at.
			
restart:
	LD IY,numstack		; Reset the stacks.

;;; This is where the fun starts.
	
constart:
	LD IX,welcome		; Display the welcome message.
	CALL i_pstring            

console:
	LD IX,prompt
	CALL i_pstring
	LD IX,strbuf		; Get a string from the user.
	CALL i_gstring

	LD IX,strbuf
	LD A,(IX+0)
	CP ':'			; A word def.
	JR Z,dodef
	CALL parse
	LD A,(exityn)
	OR A
	JR Z,console
	RET			; quit
	
;;; These are hugely inefficient. We really need a seek routine.
dodef:	INC IX
	PUSH IY
	LD IY,wordbuf
ddloop:	LD A,(IX+0)
	LD HL,lettab
	CALL inlist
	JR NZ,ddloopend
	LD (IY+0),A
	INC IX
	INC IY
	JR ddloop
ddloopend:
	XOR A
	LD (IY+0),A
	INC IX
	PUSH IX			; now points at content
redoloop:	
	;; word name now in wordbuf
	LD IY,fstore
dd2loop:
	PUSH IY
	LD IX,wordbuf
	CALL strcmp
	POP IY
	JR Z,alreadythere
	CALL seekrecord
	LD A,(IY+0)
	OR A
	JR NZ,dd2loop
	JR insertword
alreadythere:
;;; this should delete an already-there word.
;;; now go back to find the end of fstore again.
;;; IY points at the start of the existing word's name
	PUSH IY
	POP DE			; DEstination
	CALL seekrecord
	PUSH IY
	POP HL			; Source
	LD BC,0h
alreadyloop:
	LD A,(IY+0)
	OR A
	JR Z,foundend
	CALL seekrecord
	JR alreadyloop	
foundend:
	INC BC
	LDIR
	JR redoloop
insertword:
	LD IX,wordbuf
copynameloop:
	LD A,(IX+0)
	LD (IY+0),A
	INC IX
	INC IY
	OR A
	JR NZ,copynameloop
	POP IX			; at last! get the content back!
copycontentloop:	
	LD A,(IX+0)
	LD (IY+0),A
	INC IX
	INC IY
	OR A
	JR NZ,copycontentloop
	LD A,'+'
	CALL i_out	
	LD IX,wordbuf
	CALL i_pstring
	POP IY		
	JP console
		
;;; "parse" Takes the string address in IX. Reentrant.
parse:	PUSH AF
	PUSH BC
	PUSH HL
	LD A,(recurselevel)
	INC A
	LD (recurselevel),A
	CP recurselimit		; if we overrecurse
	JR NZ,ploop
ovflow:	PUSH IX
	LD IX,overflow
	CALL i_pstring
	POP IX
	JP endpse
ploop:	LD A,IYU		; *CAUTION* Undoc instruction, not portable
	CP numstack/256
	JR NZ,ovflow	; Restart upon stack over/underflow.	
	LD A,(IX+0)
	OR A			; Zero terminated: return.
	JP Z,endpse
	LD HL,numtab
	CALL inlist
	JR NZ,notnum
	CALL donumber
	JR ploop
notnum:	
	LD HL,cmdtab
findcmd:
	CP (HL)			; found it in the table?
	JR Z,foundcmd		; yes, do something with it.
	LD E,A
	LD A,(HL)
	OR A			; optimised cp 0h
	JR Z,checkf
	LD A,E
	INC HL
	INC HL
	INC HL
	JR findcmd
foundcmd:
	PUSH IX
	PUSH HL
	POP IX
	LD L,(IX+2)
	LD H,(IX+1)
	POP IX
	CALL jumphl	
nextcmd:
	INC IX
	JR ploop
checkf:	LD A,E
	LD HL,lettab
	CALL inlist
	JR NZ,nextcmd
	PUSH IY
	LD IY,wordbuf
cfloop:	LD A,(IX+0)
	LD HL,lettab
	CALL inlist
	JR NZ,cfldone
	LD (IY+0),A
	INC IX
	INC IY
	JR cfloop
cfldone:
	XOR A
	LD (IY+0),A
	POP IY
	;; the word read is now in wordbuf
	PUSH IX
	PUSH IY
	;; now find the word in fstore. You are *not* expected to understand
	;; this at all.
	LD IY,fstore
gwloop:	PUSH IY
	LD IX,wordbuf
	CALL strcmp
	POP IY
	JR Z,gotword
	CALL seekrecord
	LD A,(IY+0)
	OR A
	JR NZ,gwloop
	JR notgotword
gotword:
	CALL seekiytozero
	PUSH IY
	POP IX
	POP IY
	CALL parse
	JR gwdone
notgotword:
	LD A,'?'
	CALL i_out
	LD IX,wordbuf
	CALL i_pstring
	POP IY
gwdone:	POP IX
	DEC IX
	JR nextcmd
	
jumphl:	JP (HL)

endpse:	LD A,(recurselevel)
	DEC A
	LD (recurselevel),A
	POP HL
	POP BC
	POP AF
	RET

seekrecord:
	CALL seekiytozero
;;; FALL THROUGH!
	
;;; Really should be called seekiypastzero. But why be descriptive? :)
seekiytozero:
	LD A,(IY+0)
	INC IY
	INC BC
	OR A
	RET Z
	JR seekiytozero
	
; Constants	    12345678901234567890
welcome:	DB "zforth 0.7 - azz 98", eos
prompt:	DB cr,">",eos

cmdtab:	DB '+',cadd/256,cadd&255
	DB '-',csub/256,csub&255
	DB '*',cmul/256,cmul&255
	DB '/',cdiv/256,cdiv&255
	DB '#',cpoke/256,cpoke&255
	DB '?',cpeek/256,cpeek&255
	DB '=',equal/256,equal&255
	DB '>',more/256,more&255
	DB '<',less/256,less&255
	DB '.',outnum/256,outnum&255
	DB ',',cout/256,cout&255
	DB '^',cin/256,cin&255
	DB ']',cnor/256,cnor&255
	DB '|',cor/256,cor&255
	DB '&',cand/256,cand&255
	DB '~',cnot/256,cnot&255
	DB '%',off/256,off&255
	DB '$',double/256,double&255
	DB '@',rot/256,rot&255
	DB '!',swap/256,swap&255
	DB '\"',string/256,string&255
	DB 27h,quotechar/256,quotechar&255 ;'
	DB '(',left/256,left&255
	DB ')',right/256,right&255
	DB '}',exit/256,exit&255
	DB 0,0,0

exit:	LD A,1h
	LD (exityn),A
	RET
	
left:	CALL off
	LD A,H
	OR L
	RET NZ
	LD C,0h
leftlp:	LD A,(IX+0)
	INC IX
	CP '('
	JR NZ,lnotl
	INC C
lnotl:	CP ')'
	JR NZ,leftlp
	DEC C
	JR NZ,leftlp
	DEC IX			; as it gets incremented again... sigh
	RET

right:	LD C,0h
rightlp:
	LD A,(IX+0)
	DEC IX
	CP ')'
	JR NZ,rnotr
	INC C
rnotr:	CP '('
	JR NZ,rightlp
	DEC C
	JR NZ,rightlp
	RET
	
; if the char in A is in list HL, return Z else return NZ
inlist:	OR A			; Fix the zero-in-list bug.
	JR Z,iszero
	CP (HL)
	RET Z		; ret with Z
	LD B,A
	LD A,(HL)
	CP 0h
	JR Z,notlst
	LD A,B
	INC HL
	JR inlist
notlst:	CP 1h		; which it can't be, as it's 0h
	LD A,B
	RET		; ret with NZ
iszero:	CP 1h			; Force NZ return.
	RET
	
numtab:	DB "0123456789",0h
lettab:	DB "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ",0h
	
quotechar:
	INC IX
	LD A,(IX+0)
	LD L,A
	LD H,0h
	JR dojpon
	
cpoke:	CALL off	;:	
	EX DE,HL
	CALL off
	EX DE,HL
	PUSH IX
	PUSH HL
	POP IX
	LD (IX+0),E
	LD (IX+1),D
	POP IX
	RET

cpeek:	CALL off	;;
	PUSH IX
	PUSH HL
	POP IX
	LD L,(IX+0)
	LD H,(IX+1)
	POP IX
	JR dojpon
	
cadd:	CALL off	;+
	EX DE,HL
	CALL off
	ADD HL,DE
dojpon:	JP on

csub:	CALL off	;-
	EX DE,HL
	CALL off
	SCF
	CCF
	SBC HL,DE
	JR dojpon

more:	CALL off
	EX DE,HL
	CALL off
	SCF
	CCF
	SBC HL,DE
	JR NC,pushfalse
	JR pushtrue

less:	CALL off
	EX DE,HL
	CALL off
	SCF
	CCF
	SBC HL,DE
	JR Z,pushfalse
	JR NC,pushtrue
	JR pushfalse
	
; Unashamedly nicked from maths.z in the tetris game for z81.
cmul:	CALL off	;*
	EX DE,HL
	CALL off
	; begin borrowed code
	ld b,h
	ld c,l
	ld hl,0
	ld a,16
mulp:	and a
	rr d
	rr e
	jr nc,musp
	add hl,bc
musp:	and a
	rl c
	rl b
	dec a
	jr nz,mulp
	; end borrowed code
	JR dojpon

cdiv:	CALL off	;/
	EX DE,HL
	CALL off
	CALL divide
	JR dojpon
			
equal:	CALL off		; =
	EX DE,HL
	CALL off
	LD A,D
	CP H
	JR NZ,pushfalse
	LD A,E
	CP L
	JR NZ,pushfalse
pushtrue:	
	LD HL,0FFFFh
	JR puco
pushfalse:
	LD HL,0h
puco:	JP on

cout:	CALL off		; ,
	LD A,L
	JP i_out

cin:	CALL i_in		; ^
	LD H,0h
	LD L,A
	JP on

cnor:	CALL off		; |
	EX DE,HL
	CALL off
	LD A,H
	OR D
	CPL
	LD H,A
	LD A,L
	OR E
	CPL
	LD L,A
	JP on

cnot:	CALL off
	LD A,H
	CPL
	LD H,A
	LD A,L
	CPL
	LD L,A
	JP on

cand:	CALL off		; &
	EX DE,HL
	CALL off
	LD A,H
	AND D
	LD H,A
	LD A,L
	AND E
	LD L,A
	JP on

cor:	CALL off		; |
	EX DE,HL
	CALL off
	LD A,H
	OR D
	LD H,A
	LD A,L
	OR E
	LD L,A
	JP on

double:	CALL off
	CALL on
	JP on

swap:	CALL off
	EX DE,HL
	CALL off
	EX DE,HL
	CALL on
	EX DE,HL
	JP on

rot:	CALL off
	PUSH HL
	CALL off
	EX DE,HL
	CALL off
	EX DE,HL
	CALL on
	POP HL
	CALL on
	EX DE,HL
	JP on
		
string:	INC IX
stringl:	
	LD A,(IX+0)
	CP 34
	RET Z
	CALL i_out
	INC IX
	JR stringl
		
donumber:			; 0-9
	LD H,0h
	LD A,(IX+0)
	INC IX
	SUB '0'
	LD L,A
donextnum:
	LD A,(IX+0)
	PUSH HL
	LD HL,numtab
	CALL inlist
	POP HL
	JP NZ,on	; and thus return
	CALL multen
	LD B,0h
	SUB '0'
	LD C,A
	ADD HL,BC
	INC IX
	JR donextnum
	
; multiply the number in HL by ten
multen:	PUSH HL
	POP BC
	LD E,09h
multen0:
	ADD HL,BC
	DEC E
	JR NZ,multen0
	RET			

; push and pop, IY is DSP, HL is data
on:	LD (IY+1),H
	LD (IY+0),L
	DEC IY
	DEC IY
	RET

off:	INC IY
	INC IY
	LD H,(IY+1)
	LD L,(IY+0)
	RET

; From z81 tetris math.z
;divide, hl=hl/de, de=hl%de
;af/bc corrupt
dividey:
	dw 0
divide:
	ld b,h
	ld c,l
	;see if we're trying to divide by zero
	ld a,d
	or e
	ret z
	ld (dividey),de
	ld de,0
	ld hl,0
	ld a,16
dvlp:	push af
	and a
	rl l
	rl h
	and a
	rl e
	rl d
	bit 7,b	
	jr z,dvs1
	ld a,1
	or l
	ld l,a
dvs1:	push hl
	and a
	push de
	ld de,(dividey)
	sbc hl,de
	pop de
	jp m,dvs2
	ex (sp),hl
	ld a,1
	or e
	ld e,a
dvs2:	pop hl
	and a
	rl c
	rl b
	pop af
	dec a
	jr nz,dvlp
	ex de,hl
	ret

outnum:	CALL off
	EX DE,HL
	CALL itoa
	PUSH IX
	LD IX,numtmp
onloop:	LD A,(IX+0)
	CP ' '
	JR Z,onnext
	CP 0h
	JR Z,ondone
	CALL i_out
onnext:	INC IX
	JR onloop
ondone:	LD A,' '
	CALL i_out
	POP IX
	RET
	
;convert number in de to ascii in internal buffer
;entry: de=number
;exit:  hl=addr of number in internal buffer, NUL terminated
itoa:	ld hl,numtmp
	ld c,numtmpend-numtmp-1
itoa0:	ld a,' '
	ld (hl),a
	inc hl
	dec c
	jr nz,itoa0
	ld b,10		; Base, in theory.
	ld hl,numtmpend-1
	xor a
	ld (hl),a
dispnlp:
	push bc
	push hl
	ex de,hl
	ld e,b
	ld d,0
	call divide
	ld a,e
	add a,'0'	; this ain't no ZX81!
	ex de,hl	; so de now is result of division
	pop hl
	pop bc
	dec hl
	ld (hl),a
	ld a,d
	or e
	jp nz,dispnlp
	ret
	
numtmp: DB "      ",eos		; buffer for number output
numtmpend:	

;;; "strcmp" Compare two strings in IX and IY, and set Z if they're equal.
	
strcmp:	LD A,(IX+0)
	CP (IY+0)
	RET NZ
	OR A
	RET Z
	INC IX
	INC IY
	JR strcmp

; *** Writable data area.

overflow: DB "Stack overflow",cr,eos
hexbuf:	DB "  x", cr, eos
strbuf:	DS 80			; 80 char buffer for string input.
wordbuf: DS 40			; Buffer for the function name.
spscratch:			; Scratch space to write the SP to.
	DS 2
		
;;; The function storage area. This MUST be at the end of the program.
;;; This defines all the standard words.
fstore:
	DB "true",0h,"65535",0h
	DB "false",0h,"0",0h
	DB "t",0h,"true",0h
	DB "f",0h,"false",0h
	DB "d","up",0h,"$",0h	; Necessary because "dup" is reserved.
	DB "rot",0h,"@",0h
	DB "swap",0h,"!",0h
	DB "drop",0h,"%",0h
	DB "nip",0h,"!%",0h
	DB "over",0h,"!$@!",0h
	DB "tuck",0h,"$@!",0h
	DB "rrot",0h,"@@",0h
	DB "tdrop",0h,"!!",0h
	DB "td","up",0h,"!$@$@!",0h
	DB "L",0h,"td","up",0h
	DB "and",0h,"&",0h
	DB "or",0h,"|",0h
	DB "not",0h,"~",0h
	DB "nor",0h,"]",0h
	DB "xor",0h,"L~!~]@@]",0h;  This is wrong
	DB "peek",0h,"?255&",0h
	DB "poke",0h,"$?65280&@|!#",0h
	DB "peekw",0h,"?",0h
	DB "pokew",0h,"#",0h
	DB "v",0h,"2*7168+",0h
	DB "var",0h,"v",0h
	DB "r",0h,"v?",0h
	DB "w",0h,"v#",0h
	DB "mod",0h,"3v#$3v?/3v?*-",0h
	DB "fac",0h,"$1=~($1-fac*0)",0h
	DB "square",0h,"$*",0h
	DB "rnd",0h,"\{random}peek",0h
	DB "cr",0h,"13,",0h
	DB "getc",0h,"^",0h
	DB "putc",0h,",",0h
	DB "hello",0h,"\"hello world\"",0h
	DB "cat",0h,"1(^$,3=~)",0h
	DB "factorp",0h,"L/*=",0h
	DB "primep",0h,"1 11v#$10v#1(1-$10v?!factorp(0 11v#0)$2=~)%11v?",0h
	DB "primes",0h,"2 1(1+$.$primep.cr$30=~)",0h
	DB "bye",0h,"}",0h
	DB 0h
;;; Do not insert anything after the word table.
	END





